﻿using Microsoft.Win32;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Security;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Input;

namespace exo2csv
{
    /// <summary>
    /// 抽出結果1件分を入れるクラス
    /// </summary>
    public class ExtractedData
    {
        /// <summary>
        /// 開始位置のフレーム数（exoファイルのstart。ソートに使う）
        /// </summary>
        public int StartFrame { get; set; }

        /// <summary>
        /// 開始位置（exoファイルのstart相当だが、設定により、フレーム数表示か時間表示のどちらかが入る）
        /// </summary>
        public string Start { get; set; }

        /// <summary>
        /// レイヤ番号（exoファイルのlayer）
        /// </summary>
        public int Layer { get; set; }

        /// <summary>
        /// 抽出した文字列（exoファイルのtextを読めるようデコードしたもの）
        /// </summary>
        public string String { get; set; }
    }

    /// <summary>
    /// MainWindow.xaml の相互作用ロジック
    /// </summary>
    public partial class MainWindow : Window
    {
        /// <summary>
        /// 最後に解析完了したexoファイルのパス
        /// </summary>
        string LastExoFilePath = null;

        /// <summary>
        /// 抽出結果全件を格納するリスト
        /// </summary>
        List<ExtractedData> ExtractedDataList = new List<ExtractedData>();

        public MainWindow()
        {
            InitializeComponent();
        }

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            //ウインドウロード完了時の処理

            //DataGridに解析結果のリストをバインディング（このリストの内容が自動的にDataGridへ反映されるようになる）
            dataGridResult.ItemsSource = ExtractedDataList;

            //ウインドウタイトル変更（アプリ名とバージョン番号）
            Assembly mainAssembly = Assembly.GetEntryAssembly();
            this.Title = String.Format("exo2csv ver {0}.{1}", mainAssembly.GetName().Version.Major, mainAssembly.GetName().Version.Minor);

            //「コピーしました！」を初期非表示とする
            textBlockCopied.Visibility = Visibility.Hidden;

            //「（ここに開いたファイルの情報が出ます）」の文字はいらないので消しておく（GUI編集時に分かりやすいようにダミーで初期値を入れてあるだけ）
            textBlockFileInfo.Text = "";

            //コマンドライン引数でファイルパスを渡されていたら、自動で読み込む。
            if ((App.CommandLineArgs != null) && (App.CommandLineArgs.Length >= 1))
            {
                //解析して結果を表示
                LoadAndAnalyzeExo(App.CommandLineArgs[0]);  //複数ファイル渡された場合、先頭の引数で渡されたパスを読み込む
            }
        }

        private void Window_PreviewDragOver(object sender, DragEventArgs e)
        {
            //ドラッグ中のカーソルがフォームに入ったときの処理

            //ファイルがドラッグされたらカーソル変更
            if (e.Data.GetDataPresent(DataFormats.FileDrop, true))
            {
                e.Effects = DragDropEffects.Copy;   //カーソルを変更
            }
            else
            {
                e.Effects = DragDropEffects.None;   //カーソルを通常に戻す
            }
            e.Handled = true;
        }

        private void Window_PreviewDrop(object sender, DragEventArgs e)
        {
            //ドロップされたときの処理

            //ドロップされたファイルのファイルパスを取得
            var dropFiles = e.Data.GetData(DataFormats.FileDrop) as string[];
            if (dropFiles == null) return;  //ドロップされたのがファイルでないなら何もせず返す

            //解析して結果を表示　※この処理が終わるまでドラッグ元のエクスプローラもフリーズするが、短時間なので特に対策しない
            LoadAndAnalyzeExo(dropFiles[0]);  //複数ファイルドロップされた場合でも、読み込むのは先頭の1つだけ
        }

        private void buttonOpenExoFile_Click(object sender, RoutedEventArgs e)
        {
            //「exoファイルを開く...」ボタンが押された時の処理

            //ダイアログのインスタンスを生成
            var dlg = new OpenFileDialog();

            //ダイアログの設定
            dlg.DefaultExt = ".exo"; //デフォルト拡張子
            dlg.Filter = "AviUtl拡張編集 ObjectFile|*.exo|全てのファイル (*.*)|*.*";

            //ダイアログを表示する
            Nullable<bool> result = dlg.ShowDialog();

            //ダイアログ操作の結果を確認
            if (result == true)
            {
                //正しくファイル選択されたので、開いて解析
                LoadAndAnalyzeExo(dlg.FileName);
            }
        }

        /// <summary>
        /// exoファイルを読み込み、解析して、結果を表示する
        /// </summary>
        /// <param name="filePath"></param>
        private void LoadAndAnalyzeExo(string filePath)
        {
            //解析結果欄の上にファイル名表示（読み込んだが1行もデータが無かった場合などに無反応に見えないよう、読み込み失敗の場合でもファイル名だけは出す）
            textBlockFileInfo.Text = Path.GetFileName(filePath);

            //ファイルが存在しなければエラーを出す
            if (File.Exists(filePath) == false)
            {
                MessageBox.Show("エラー： *.exo ファイルが存在しません。読み込み処理を中断します。filePath=" + filePath);
                return;
            }

            //DataGridの各列幅を初期化（こうしておかないと、セル内容が短くなっても列幅が広がったまま戻らないので）
            foreach (var column in dataGridResult.Columns)
            {
                column.Width = System.Windows.Controls.DataGridLength.SizeToCells;
                column.Width = System.Windows.Controls.DataGridLength.SizeToHeader;
                column.Width = System.Windows.Controls.DataGridLength.Auto;
            }

            //前回読み出したデータをクリア
            ExtractedDataList.Clear();

            //ここから解析開始。

            //ファイルを1行ずつ読みながら、フレームレート情報を取り出していく
            double rate;    //exoファイルのrate（AviUtl拡張編集が出力したexoファイルでは必ず整数になるが、外部ツールから出力したexoファイルでは小数になる場合があるようなので、doubleにする）
            double scale;   //exoファイルのscale（rateに合わせてdoubleにする）
            bool hasFrameRate = AnalyzeFrameRate(filePath, out rate, out scale);    //falseが返ってきたら、exoファイルの先頭が欠けているなどの理由でフレームレート情報を取得できなかった状態

            //ファイルを1行ずつ読みながら、各テキストオブジェクトの開始位置、レイヤ番号、文字列を取り出していく
            AnalyzeTextObject(filePath, hasFrameRate, rate, scale);

            //ここまでくれば解析完了している。

            //Listを開始位置とレイヤ番号の昇順でソート
            ExtractedDataList.Sort((a, b) => (((a.StartFrame - b.StartFrame) != 0) ? (a.StartFrame - b.StartFrame) : (a.Layer - b.Layer)));   //開始位置が違うなら開始位置でソート、同じならレイヤ番号でソート

            //DataGridの表示を強制更新（こうしないと、スクロールなどするまで更新されないので）
            dataGridResult.Items.Refresh();

            //ファイル名を覚えておく（再解析や、名前を付けて保存する時のデフォルトファイル名として使う）
            LastExoFilePath = filePath;

            //読み込み前に無効としていたコントロールを有効化
            buttonSaveToFile.IsEnabled = true;
            buttonCoypToClipboard.IsEnabled = true;

            //フレームレート情報を読み込めなかった場合、フレーム数を時間換算できていないため警告表示する
            if ((checkBoxStartAsFrame.IsChecked == false) && (hasFrameRate == false))
            {
                //警告表示
                MessageBox.Show("警告：解析は完了しましたが、フレームレート情報を読み取れなかったため、開始位置を時間ではなくフレーム数で表示します。 *.exo ファイルが不完全ではないかご確認ください。");
            }
        }

        /// <summary>
        /// ファイルを1行ずつ読みながら、フレームレート情報（rateとscale）を取り出す
        /// </summary>
        /// <param name="filePath"></param>
        /// <param name="rate">exoファイルのrate</param>
        /// <param name="scale">exoファイルのscale</param>
        /// <returns>false:フレームレート情報が読めなかった true:エラーなし（フレームレート情報を読めた）</returns>
        private bool AnalyzeFrameRate(string filePath, out double rate, out double scale)
        {
            rate = -1;   //最初はあり得ない値にしておく
            scale = -1;   //最初はあり得ない値にしておく

            using (var sr = new StreamReader(filePath, Encoding.GetEncoding("Shift_JIS")))
            {
                string line;
                while ((line = sr.ReadLine()) != null)
                {
                    //1行分の解析

                    //※rate→scaleの順で並んでいるはずだが、念のため、順序が逆になっていても大丈夫なようにしている

                    //フレームレートかどうか確認
                    string keyword = "rate=";
                    if (line.StartsWith(keyword))
                    {
                        //フレームレートを見付けたので記憶
                        string value = line.Substring(keyword.Length, line.Length - keyword.Length);  //値を取り出す
                        rate = double.Parse(value); //数値に変換して記憶　※一部の外部ツールが出力したexoファイルでは小数になるようなので、doubleでPerseする

                        //トラブルシューティング用に表示しておく
                        textBlockFileInfo.Text += ", rate=" + rate;
                    }

                    //スケールかどうか確認
                    keyword = "scale=";
                    if (line.StartsWith(keyword))
                    {
                        //スケールを見付けたので記憶
                        string value = line.Substring(keyword.Length, line.Length - keyword.Length);  //値を取り出す
                        scale = double.Parse(value); //数値に変換して記憶

                        //トラブルシューティング用に表示しておく
                        textBlockFileInfo.Text += ", scale=" + scale;
                    }

                    //終了判定
                    if ((rate != -1) && (scale != -1))
                    {
                        //フレームレート情報を読み出せたので終了
                        return true;
                    }

                    //次のオブジェクトの開始かどうか確認
                    if (Regex.IsMatch(line, @"\[\d+\]"))
                    {
                        // "[?]" （?部分は数字）という形式なので、ここから先は次のオブジェクト
                        return false;    //フレームレート情報が読めなかったということなので、エラーを返す
                    }
                }
            }

            //ファイル末尾までいった場合、ファイル先頭が欠落しているなどでフレームレート情報が得られなかった状況なのでエラーを返す
            return false;
        }

        /// <summary>
        /// ファイルを1行ずつ読みながら、各オブジェクトの開始位置、レイヤ番号、文字列を取り出していく
        /// </summary>
        /// <param name="filePath"></param>
        /// <param name="hasFrameRate"></param>
        /// <param name="rate"></param>
        /// <param name="scale"></param>
        private void AnalyzeTextObject(string filePath, bool hasFrameRate, double rate, double scale)
        {
            using (var sr = new StreamReader(filePath, Encoding.GetEncoding("Shift_JIS")))
            {
                int phase = 0;  //解析処理の状態 0:次のオブジェクトまでスキップ中 / 1-2:開始位置を検索中 / 3:テキストデータを検索中
                var extractedData = new ExtractedData();  //1オブジェクト分のデータを一時的に入れる変数

                string line;
                while ((line = sr.ReadLine()) != null)
                {
                    //1行分の解析

                    //現在の状態により分岐
                    switch (phase)
                    {
                        case 0: //0:次のオブジェクトまでスキップ中

                            //次のオブジェクトの開始かどうか確認
                            if (Regex.IsMatch(line, @"\[\d+\]"))
                            {
                                // "[?]" （?部分は数字）という形式なので、ここから先は次のオブジェクト
                                extractedData = new ExtractedData();    //新しいデータを入れる準備

                                //次のphaseへ
                                ++phase;
                            }

                            break;

                        case 1: //1-2:開始位置とレイヤを検索中
                                //go through
                        case 2: //1-2:開始位置とレイヤを検索中
                            //※開始位置→レイヤの順で並んでいるはずだが、念のため、開始位置とレイヤの順序が逆になっていても大丈夫なようにしている

                            //開始位置かどうか確認
                            string keyword = "start=";
                            if (line.StartsWith(keyword))
                            {
                                //開始位置を見付けたので記憶
                                string value = line.Substring(keyword.Length, line.Length - keyword.Length);  //値を取り出す
                                extractedData.StartFrame = int.Parse(value);    //数値に変換して記憶

                                //開始位置をfpsに変換する設定かどうか、もしくはフレームレート情報をもっていないかどうか（その場合はそもそも換算できない）を確認
                                if ((checkBoxStartAsFrame.IsChecked == true) || (hasFrameRate == false))
                                {
                                    //フレーム数のまま
                                    extractedData.Start = extractedData.StartFrame.ToString();
                                }
                                else
                                {
                                    //時間に換算する
                                    double fps = rate / scale;  //フレーム毎秒
                                    int hour = (int)((extractedData.StartFrame - 1) / fps / 3600 % 60);  //時
                                    int min = (int)((extractedData.StartFrame - 1) / fps / 60 % 60); //分
                                    int sec = (int)((extractedData.StartFrame - 1) / fps % 60);  //秒
                                    int ms = (int)((extractedData.StartFrame - 1) % fps * 100 / fps);    //0.01秒
                                    extractedData.Start = @"#" + String.Format("{0:D2}:{1:D2}:{2:D2}.{3:D2}", hour, min, sec, ms);  //先頭に # を付加するのは、CSVをExcelで開いたときに時刻と解釈されて勝手に変換されるのを防ぐため
                                }

                                //次のphaseへ
                                ++phase;
                            }

                            //レイヤ番号かどうか確認
                            keyword = "layer=";
                            if (line.StartsWith(keyword))
                            {
                                //レイヤ番号を見付けたので記憶
                                string value = line.Substring(keyword.Length, line.Length - keyword.Length);  //値を取り出す
                                extractedData.Layer = int.Parse(value); //数値に変換して記憶

                                //次のphaseへ
                                ++phase;
                            }

                            break;

                        case 3: //3:テキストデータを検索中

                            //テキストオブジェクトのテキストデータかどうかを確認
                            keyword = textBoxTextDataKeyword.Text;  //基本的には「text=」だが、汎用性を考えて設定で変更可能にしている（例えば「file=」にするなど用途がありそう）
                            if (line.StartsWith(keyword))
                            {
                                //テキストデータなので、値を取り出して記憶する

                                //値を取り出す
                                line = line.Substring(keyword.Length);  //値を取り出す
                                if (checkBoxEnableDecode.IsChecked == true)
                                {
                                    line = DecodeExoText(line); //そのままでは読めない形式になっているのでデコード
                                }

                                //テキストデータ先頭が除外キーワードと一致するならスキップ（主にPSDToolKitの「PSDファイルオブジェクト」、「字幕表示オブジェクト」といった、テキストデータ部分にスクリプトが記述されたテキストオブジェクトを除くため）
                                bool isSkip = false;
                                foreach (var word in textBoxTextDataNgWord.Text.Split(';'))
                                {
                                    if ((word != "") && line.StartsWith(word))   //除外キーワード欄の末尾がセミコロンの場合、wordが最後に空文字になるので、その場合は判定しない
                                    {
                                        //一致したので、このテキストデータはスキップする
                                        isSkip = true;
                                        break;
                                    }
                                }

                                if (isSkip == false)
                                {
                                    //スキップしないテキストデータなので処理続行

                                    //テキストデータをさらに1行ずつに分解して、必要なところだけ取り出す
                                    extractedData.String = "";  //このあと文字を後ろに足していくので、初期値として空文字を入れておく
                                    using (var stream = new StringReader(line))
                                    {
                                        //テキストデータ内の1行分の解析
                                        string textDataLine;
                                        while ((textDataLine = stream.ReadLine()) != null)  //AviUtl拡張編集は改行コードを "\r\n" で出力するが、外部ツールでは "\n" で出力する場合もあるようだ。しかしこの処理だと両方に対応できる（混在していてもOK）
                                        {
                                            //テキストデータ内の1行分の先頭が除外キーワードと一致するならスキップ（主にPSDToolKitの「字幕準備オブジェクト」、「統合準備オブジェクト」の不要部分のみ除くため）
                                            isSkip = false;
                                            foreach (var word in textBoxTextDataLineNgWord.Text.Split(';'))
                                            {
                                                if ((word != "") && textDataLine.StartsWith(word))   //各行除外キーワード欄の末尾がセミコロンの場合、wordが最後に空文字になるので、その場合は判定しない
                                                {
                                                    //一致したので、この行はスキップする
                                                    isSkip = true;
                                                    break;
                                                }
                                            }

                                            if (isSkip == false)
                                            {
                                                //スキップしない行なので記憶
                                                //（本当は最後に、PSDToolKit対応として ']==].."]==]"..[==[' を ']==]' に置き換える必要があるが、稀なので省略）
                                                extractedData.String += textDataLine + @"\r\n"; //1行分追記
                                            }
                                        }
                                    }

                                    //最後が改行の場合は消しておく（PSDToolKitの字幕準備の場合に、最後が必ず改行が付いてしまうので。手動で末尾に付けた改行も消えてしまうが仕方なしとする）
                                    string str = @"\r\n";
                                    if (extractedData.String.EndsWith(str))
                                    {
                                        extractedData.String = extractedData.String.Remove(extractedData.String.Length - str.Length, str.Length);
                                    }

                                    //改行を"\r\n"に置換する設定かどうか確認
                                    if (checkBoxEnableReturn.IsChecked == false)
                                    {
                                        //置換する
                                        extractedData.String = extractedData.String.Replace(@"\r\n", Environment.NewLine);
                                    }

                                    //ここまでくれば、今のテキストオブジェクトからの値の取得が完了している。

                                    //必要な情報が集まったので、今回のデータをListに保管
                                    ExtractedDataList.Add(extractedData);

                                    //このオブジェクトの取得は終わったので、最初のphase（次のオブジェクト開始待ち）へ
                                    phase = 0;
                                }
                            }

                            //次のオブジェクトの開始かどうか確認
                            if (Regex.IsMatch(line, @"\[\d+\]"))
                            {
                                // "[?]" （?部分は数字）という形式なので、ここから先は次のオブジェクト

                                //次のオブジェクトに到達した場合、今のオブジェクトはテキストオブジェクトではなかったということなので、破棄
                                extractedData = new ExtractedData();    //新しいデータを入れる準備

                                //phase1（開始位置の検索）へ
                                phase = 1;
                            }

                            break;
                    }
                }
            }
        }

        /// <summary>
        /// exoファイルのテキストオブジェクトに含まれるテキストデータを通常の文字列にデコードする
        /// </summary>
        /// <param name="text"></param>
        /// <returns></returns>
        private string DecodeExoText(string text)
        {
            //参考：AviUtlのexoファイル出力時のテキストの変換 - Qiita
            //　　　https://qiita.com/rrryutaro/items/57ad42b7b66882c39797
            //　　　（この記事で紹介されているのと逆の変換をすればよい）

            //渡された文字列がそもそもexoファイルのテキストオブジェクトでないと思われる場合（16進数の数字以外を含むか「0000」を含まない場合）、エラーメッセージを返す（これをそのまま表示してもらう想定）
            if ((text.All(c => Uri.IsHexDigit(c)) == false) || (text.Contains("0000") == false))
            {
                return "エラー：テキストデータのデコードに失敗。「テキストデータをデコード」のチェックを外してください。";
            }

            //16進数文字列を4文字ずつ見ていき"0000"以降を削除（"0000"以降にも何か文字列がある場合があるが、AviUtl上でテキストを編集して短くした場合に残るゴミと思われるため無視）
            string str = text;  //最後まで "0000" が見つからなかった場合は全体を使う
            for (int i = 0; i < text.Length; i += 4)
            {
                if (text.Substring(i, 4) == "0000")
                {
                    //iからの4文字が "0000" なので、先頭からiの前までが必要な文字列
                    str = text.Substring(0, i);
                    break;
                }
            }

            //16進数文字列をbyteのListに変換　参考：https://docs.microsoft.com/ja-jp/dotnet/csharp/programming-guide/types/how-to-convert-between-hexadecimal-strings-and-numeric-types
            var byteList = new List<byte>();
            for (int i = 0; i < str.Length / 2; ++i)
            {
                //2文字取り出す
                string hexStr = str.Substring(i * 2, 2);

                //byteに変換してリストに入れる
                byteList.Add(Convert.ToByte(hexStr, 16));
            }

            //byteのListをUTF-16エンコードされた文字列と見做してデコード
            string result = Encoding.Unicode.GetString(byteList.ToArray());

            return result;
        }

        private void checkBoxCommon_Checked(object sender, RoutedEventArgs e)
        {
            //いずれかのチェックボックスを変更（チェック入れたor外した）したときの処理
            ReloadExo();    //新しい設定で再読込
        }

        private void buttonReloadExoFile_Click(object sender, RoutedEventArgs e)
        {
            //いずれかの「適用」ボタンを押した時の処理
            //（TextBoxを1文字変更するだけで再読み込みするのはさすがに重いので、別にボタンを設けた）
            ReloadExo();  //新しい設定で再読込
        }

        private void textBoxCommon_KeyDown(object sender, System.Windows.Input.KeyEventArgs e)
        {
            //TextBoxでキー入力があったときの共通処理
            //エンターキーが押されたら「適用」ボタンを押した時と同じ処理をする
            if (e.Key == Key.Return)
            {
                ReloadExo();  //新しい設定で再読込
            }
        }

        /// <summary>
        /// 最後に開いたexoファイルを再読み込みする
        /// </summary>
        private void ReloadExo()
        {
            if (LastExoFilePath != null)
            {
                //一度でもファイルを開いたことがあるので、自動で再解析する
                LoadAndAnalyzeExo(LastExoFilePath);   //最後に開いたファイルを再解析
            }
        }

        private void buttonSaveToFile_Click(object sender, RoutedEventArgs e)
        {
            //「名前を付けて保存...」ボタンが押された時の処理

            //ダイアログのインスタンスを生成
            var dlg = new SaveFileDialog();

            //ダイアログの設定
            dlg.FileName = Path.GetFileNameWithoutExtension(LastExoFilePath) + ".csv"; //デフォルトファイル名
            dlg.DefaultExt = ".csv"; //デフォルト拡張子
            dlg.Filter = "CSV ファイル|*.csv|全てのファイル (*.*)|*.*";
            dlg.InitialDirectory = Path.GetDirectoryName(LastExoFilePath);  //最後に開いたファイルがあるフォルダを初期ディレクトリにする

            //ダイアログを表示する
            Nullable<bool> result = dlg.ShowDialog();

            //ダイアログ操作の結果を確認
            if (result == true)
            {
                //正しくファイル選択されたので、名前を付けて保存
                SaveResult(dlg.FileName);
            }
        }

        /// <summary>
        /// 解析結果をファイルに保存する
        /// </summary>
        /// <param name="filePath"></param>
        private void SaveResult(string filePath)
        {
            //保存先フォルダがなければエラーを出す
            string folderPath = Path.GetDirectoryName(filePath);
            if ((folderPath != "") && (Directory.Exists(folderPath) == false))   //フォルダパスがない場合は、カレントディレクトリに保存するのでエラーにしない（現状無意味だが、将来コマンドライン引数で保存ファイル名を指定できるようにした時に備えている）
            {
                MessageBox.Show("エラー：保存先フォルダが存在しません。保存処理を中断します。filePath=" + filePath);
                return;
            }

            //エンコード設定を確認
            Encoding encoding = Encoding.UTF8;  //デフォルト（UTF-8設定） ※SJISだとExcelで開いたとき文字化けしやすい気がするため、デフォルトをUTF8にした
            if (comboBoxEncode.SelectedIndex == 1)
            {
                encoding = Encoding.GetEncoding("shift-jis");   //Shift-JIS設定
            }

            //保存実行
            try
            {
                File.WriteAllText(filePath, GenerateCsv(), encoding);   //CSV形式のテキストを生成して、ファイルに保存
            }
            catch (UnauthorizedAccessException ex)
            {
                MessageBox.Show("エラー：ファイルへの保存に失敗しました。保存先ファイルが読み取り専用になっていないかご確認ください。message=" + ex.Message);
                return;
            }
            catch (SecurityException ex)
            {
                MessageBox.Show("エラー：ファイルへの保存に失敗しました。アクセス許可がありません。message=" + ex.Message);
                return;
            }
            catch (IOException ex)
            {
                MessageBox.Show("エラー：ファイルへの保存に失敗しました。他のソフトで保存先ファイルを開いていないかご確認ください。message=" + ex.Message);
                return;
            }
            catch (Exception ex)
            {
                MessageBox.Show("エラー：ファイルへの保存に失敗しました。message=" + ex.ToString()); //想定外の例外なので、詳細な例外情報を出す
                return;
            }
        }

        private void buttonCoypToClipboard_Click(object sender, RoutedEventArgs e)
        {
            //「クリップボードへコピー」ボタンが押された時の処理

            Clipboard.SetText(GenerateCsv());  //CSV形式のテキストを生成して、クリップボードへコピー

            //反応がないと分かりづらいので、「コピーしました！」と表示
            textBlockCopied.Visibility = Visibility.Visible;

            //1秒後に消す（非同期処理）　※1秒以内に複数回ボタンを押すと表示がおかしくなるが、さほど支障ないので対策しない
            var taskHiddenCopiedTextBlock = Task.Run(() =>
            {
                //1秒待つ
                Task.Delay(1000).Wait();

                //UIスレッド直接触れないのでInvokeする
                this.Dispatcher.Invoke((Action)(() =>
                {
                    textBlockCopied.Visibility = Visibility.Hidden;
                }));
            });
        }

        /// <summary>
        /// 解析結果からCSV形式テキストを作る
        /// </summary>
        /// <returns></returns>
        private string GenerateCsv()
        {
            var result = new StringBuilder();

            using (var sw = new StringWriter(result))
            {
                sw.WriteLine("開始位置,レイヤ,抽出テキスト");    //タイトル行

                foreach (var data in ExtractedDataList)
                {
                    //1データ分を、CSVファイルの1行分に変換

                    //CSVでは、テキストはダブルクオーテーションで囲む（そうしないと改行混じりの場合やカンマを含む場合にExcelで正しく扱えないので）
                    string text = data.String.Replace("\"", "\"\"");  //元々あるダブルクオーテーションは、ダブルクオーテーション2つに置換（RFC4180準拠のエスケープ。Excelならこれで大丈夫だが、処理系によっては正しく扱えない場合があるかも）
                    text = "\"" + text + "\"";  //ダブルクオーテーションで囲む

                    //1行書き込む
                    sw.WriteLine("{0},{1},{2}", data.Start, data.Layer, text);
                }
            }

            return result.ToString();
        }
    }
}
